/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "my-string.h"

#include "config.h"
#include "db.h"
#include "execute.h"
#include "functions.h"
#include "list.h"
#include "log.h"
#include "match.h"
#include "parse_cmd.h"
#include "parser.h"
#include "storage.h"
#include "unparse.h"
#include "utils.h"

/* internals */

struct parser_state {
    Var		code;		/* a list of strings */
    int		cur_string;	/* which string? */
    int		cur_char;	/* which character in that string? */
    Var		errors;		/* a list of strings */
};

static void
my_error(void *data, const char *msg)
{
    struct parser_state	*state = (struct parser_state *) data;
    Var			v;
    
    v.type = TYPE_STR;
    v.v.str = str_dup(msg);
    state->errors = listappend(state->errors, v);
}

static int
my_getc(void *data)
{
    struct parser_state	*state = (struct parser_state *) data;
    Var			code;
    char		c;
    
    code = state->code;
    if (task_timed_out  ||  state->cur_string > code.v.list[0].v.num)
	return EOF;
    else if (!(c = code.v.list[state->cur_string].v.str[state->cur_char])) {
	state->cur_string++;
	state->cur_char = 0;
	return '\n';
    } else {
	state->cur_char++;
	return c;
    }
}

static void
my_ungetc(void *data, char c)
{
    struct parser_state	*state = (struct parser_state *) data;
    Var			code;
    
    code = state->code;
    if (state->cur_string == 1  &&  state->cur_char == 0)
	errlogf("EVAL's UNGETC: At start of code\n");
    else {
	char	cc;
	
	if (state->cur_char > 0)
	    state->cur_char--;
	else {
	    state->cur_string--;
	    state->cur_char = strlen(code.v.list[state->cur_string].v.str);
	}
	cc = code.v.list[state->cur_string].v.str[state->cur_char];
	if (!(c == cc || (cc == '\0' && c == '\n')))
	    errlogf("EVAL's UNGETC: Wrong char; c = %d and cc = %d\n", c, cc);
    }
}

Parser_Client list_parser_client = { my_error, my_ungetc, my_getc };

/* externals */

Program *
parse_list_as_program(Var code, Var *errors)
{
    struct parser_state	state;
    Program	       *program;
    
    state.code = code;
    state.cur_string = 1;
    state.cur_char = 0;
    state.errors = new_list(0);
    program = parse_program(list_parser_client, &state);
    *errors = state.errors;

    return program;
}

static package
bf_verbs(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object) */
    Var r;

    r = get_verb_list(arglist.v.list[1].v.obj, progr);
    free_var(arglist);
    if (r.type == TYPE_ERR)
	return make_error_pack(r.v.err);
    else
	return make_var_pack(r);
}

static package
bf_verb_info(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name) */
    Var r;

    r = get_verb_info(arglist.v.list[1].v.obj, arglist.v.list[2].v.str, progr);
    free_var(arglist);
    if (r.type == TYPE_ERR)
	return make_error_pack(r.v.err);
    else
	return make_var_pack(r);
}

static package
bf_set_verb_info(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name, {owner, perms, name}) */
    enum error e = set_verb_info(arglist.v.list[1].v.obj,
				 arglist.v.list[2].v.str,
				 progr, arglist.v.list[3]);
    free_var(arglist);
    if (e != E_NONE)
	return make_error_pack(e);
    else
	return no_var_pack();
}

static package
bf_verb_args(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name) */
    Var r;

    r = get_verb_args(arglist.v.list[1].v.obj, arglist.v.list[2].v.str,
		      progr);
    free_var(arglist);
    if (r.type == TYPE_ERR)
	return make_error_pack(r.v.err);
    else
	return make_var_pack(r);
}

static package
bf_set_verb_args(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name, {dobj, prep, iobj}) */
    enum error e = set_verb_args(arglist.v.list[1].v.obj,
				 arglist.v.list[2].v.str, progr,
				 arglist.v.list[3]);
    free_var(arglist);
    if (e != E_NONE)
	return make_error_pack(e);
    else
	return no_var_pack();
}

static package
bf_add_verb(Var arglist, Byte next, void *vdata, Objid progr) 
{ /* (object, info, args) */
    enum error e = add_verb(arglist.v.list[1].v.obj, progr, arglist.v.list[2],
			    arglist.v.list[3]);
    free_var(arglist);
    if (e != E_NONE)
	return make_error_pack(e);
    else
	return no_var_pack();
}

static package
bf_delete_verb(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name) */
    enum error e = delete_verb(arglist.v.list[1].v.obj,
			       arglist.v.list[2].v.str, progr);

    free_var(arglist);
    if (e != E_NONE)
	return make_error_pack(e);
    else
	return no_var_pack();
}

static package
bf_verb_code(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name [, fully-paren [, indent]]) */
    Var 	r;
    int		parens = 0;
    int 	indent = 1;
    enum error 	e;    

    if (arglist.v.list[0].v.num >= 3)
	parens = arglist.v.list[3].v.num;
    if (arglist.v.list[0].v.num >= 4)
	indent = arglist.v.list[4].v.num;

    e = get_verb_code(arglist.v.list[1].v.obj, arglist.v.list[2].v.str,
		      progr, &r, parens, indent);
    free_var(arglist);
    if (e != E_NONE)
	return make_error_pack(e);
    else
	return make_var_pack(r);
}

static package
bf_set_verb_code(Var arglist, Byte next, void *vdata, Objid progr) 
{ /* (object, vname, code) */
    Var r;

    r = set_verb_code(arglist.v.list[1].v.obj, arglist.v.list[2].v.str,
		      progr, arglist.v.list[3]);
    free_var(arglist);
    if (r.type == TYPE_ERR)
	return make_error_pack(r.v.err);
    else
	return make_var_pack(r);
}

static package
bf_eval(Var arglist, Byte next, void *data, Objid progr)
{
    package p;
    if (next == 1) {
	
	if (!is_programmer(progr)) {
	    free_var(arglist);
	    p = make_error_pack(E_PERM);
	} else {
	    Var		errors;
	    Program    *program = parse_list_as_program(arglist, &errors);
	    
	    free_var(arglist);
	    if (program) {
		free_var(errors);
		if (setup_activ_for_eval(program))
		    p = make_call_pack(2, 0);
		else {
		    free_program(program);
		    p = make_error_pack(E_MAXREC);
		}
	    } else {
		Var r;
		
		r = new_list(2);
		r.v.list[1].type = TYPE_NUM;
		r.v.list[1].v.num = 0;
		r.v.list[2] = errors;
		p = make_var_pack(r);
	    }
	}
    } else { /* next == 2 */
	Var r;
	
	r = new_list(2);
	r.v.list[1].type = TYPE_NUM;
	r.v.list[1].v.num = 1;
	r.v.list[2] = arglist; 
	p = make_var_pack(r);
    }
    return p;
}

/* register */

void
register_verbs(void)
{ 
    (void) register_function("verbs", 1, 1, bf_verbs, TYPE_OBJ);
    (void) register_function("verb_info", 2, 2, bf_verb_info,
			     TYPE_OBJ, TYPE_STR);
    (void) register_function("set_verb_info", 3, 3, bf_set_verb_info,
			     TYPE_OBJ, TYPE_STR, TYPE_LIST);
    (void) register_function("verb_args", 2, 2, bf_verb_args,
			     TYPE_OBJ, TYPE_STR);
    (void) register_function("set_verb_args", 3, 3, bf_set_verb_args,
			     TYPE_OBJ, TYPE_STR, TYPE_LIST);
    (void) register_function("add_verb", 3, 3, bf_add_verb,
			     TYPE_OBJ, TYPE_LIST, TYPE_LIST);
    (void) register_function("delete_verb", 2, 2, bf_delete_verb,
			     TYPE_OBJ, TYPE_STR);
    (void) register_function("verb_code", 2, 4, bf_verb_code,
			     TYPE_OBJ, TYPE_STR, TYPE_NUM, TYPE_NUM);
    (void) register_function("set_verb_code", 3, 3, bf_set_verb_code,
			     TYPE_OBJ, TYPE_STR, TYPE_LIST);
    (void) register_function("eval", 1, 1, bf_eval, TYPE_STR);
}

char rcsid_verbs[] = "$Id: verbs.c,v 1.16 1992/10/23 23:03:47 pavel Exp $";

/* $Log: verbs.c,v $
 * Revision 1.16  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.15  1992/10/23  22:23:43  pavel
 * Eliminated all uses of the useless macro NULL.
 *
 * Revision 1.14  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.13  1992/10/17  20:58:25  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 *
 * Revision 1.12  1992/10/06  18:25:57  pavel
 * Changed name of global Parser_Client to avoid a name clash.
 *
 * Revision 1.11  1992/09/24  16:44:38  pavel
 * Made the parsing done by eval() and set_verb_code() abort if the task runs
 * out of seconds.
 *
 * Revision 1.10  1992/09/21  17:38:05  pavel
 * Restored RCS log information.
 *
 * Revision 1.9  1992/09/14  18:41:15  pjames
 * Moved rcsid to bottom.
 * Changed bf_eval to avoid a compiler warning when compiled un-optimized.
 *
 * Revision 1.8  1992/09/14  17:31:23  pjames
 * Moved db_modification code to db modules.
 *
 * Revision 1.7  1992/09/08  22:00:18  pjames
 * Renambed bf_verb.c to verbs.c
 *
 * Revision 1.6  1992/08/31  22:29:24  pjames
 * Changed some `char *'s to `const char *'
 *
 * Revision 1.5  1992/08/28  16:16:14  pjames
 * Changed myfree(*, M_STRING) to free_str(*).
 * Changed some strref's to strdup.
 *
 * Revision 1.4  1992/08/21  00:42:59  pavel
 * Renamed include file "parse_command.h" to "parse_cmd.h".
 *
 * Revision 1.3  1992/08/10  17:21:32  pjames
 * Updated #includes.  Updated to use new registration format.  Built in
 * functions now only receive programmer, instead of entire Parse_Info.
 *
 * Revision 1.2  1992/07/20  23:55:24  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
